/* Copyright (c) Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2018 All Rights Reserved
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * This is the hostinglimits module for apache 2.X
 * author Igor Seletskiy <iseletsk@cloudlinux.com>
 * author Alexey Berezhok <alexey.berezhok@cloudlinux.com>
 * author Anton Volkov <avolkov@cloudlinux.com>
 * author Eugen Vodilov <evodilov@cloudlinux.com>
 * 
 */

/*
 * Set next macro-names:
 * APACHE2_2 - code for apache 2.2
 * APACHE2_0 - code for apache 2.0
 */

#ifndef APACHE2_2
#ifndef APACHE2_0
#error "This source is for Apache version 2.0 or 2.2 only"
#endif
#endif

#include "mod_hostinglimits_version.h"
#define MOD_HOSTINGLIMITS_SIGNAL "/var/run/modhostinglimit.flag"
#define LVE_MIN_UID 500
#define X_LVE_ID "X-LVE-ID"

#define SUPHP_MODULE "mod_suphp.c"

#include "apr_general.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "apr_thread_proc.h"
#include "ap_config_auto.h"
#include "apr_hash.h"
#include "apr_tables.h"

#define CORE_PRIVATE

#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"
#include "mod_core.h"
#include <unistd.h>
#include "unixd.h"
#include "mpm_common.h"

#include "lve/lve-ctl.h"

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>
#include <stdint.h>

#include <sys/types.h>

#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <sched.h>
#include <linux/unistd.h>
#include <sys/syscall.h>
#include <errno.h>
#include <dlfcn.h>
#include <string.h>

#include <link.h>
#include <limits.h>

#if AP_MODULE_MAGIC_AT_LEAST(20050127, 0)
#include "ap_regex.h"
#else
#ifndef _PCREPOSIX_H
#include "regex.h"
#endif

#define ap_regex_t regex_t
#define ap_regmatch_t regmatch_t

#define ap_regexec regexec
#define ap_regcomp regcomp
#define ap_regfree regfree

#define AP_REG_EXTENDED REG_EXTENDED
#define AP_REG_NOSUB REG_NOSUB
#endif

#define PREFIX "mod_hostinglimits: "

#define SECURE_LINKS_MIN_UID 100

module AP_MODULE_DECLARE_DATA hostinglimits_module;

// HTTP error code to return to client if resources limit is reached AND LVEErrorCode directive is not present in module configuration
#define DEFAULT_HTTP_ERROR_CODE 508
#define DEFAULT_HTTP_ERROR_CODE_NEW 429

// HTTP error code should be in these limits
#define MAX_HTTP_SERVER_ERROR_CODE 599
#define MIN_HTTP_SERVER_ERROR_CODE 400

//Максимальная длина URL
#define MAX_ERROR_URL_LEN 128

#define X_LVE_ID_HEADER "X_LVE_ID_HEADER"

// configuration data
typedef struct {
  unsigned int skip;
  unsigned int secure_links;
  uint32_t lve_id;
  apr_array_header_t *allowed_handlers;	// A list of handlers which will be put in LVE
  apr_array_header_t *denied_handlers;	// A list of handlers which will be NOT put in LVE
  unsigned int http_error_code;	// Integer HTTP error code to return to client if resources limit is reached
  int err_doc_found;		// true = "ErrorDocument 508" directive is found in Apache configuration
  const char *lve_user;
  apr_array_header_t *debug_sites;
  apr_array_header_t *debug_uids;
  uid_t ruid_uid;
  uid_t itk_uid;
  gid_t ruid_gid;
  gid_t itk_gid;
  uint32_t retryAfter;
  int mode;
  char *header;
  const char *path_regex;	// Compiled regex buffer
  uid_t suphp_uid;
  uid_t suphp_gid;
  int use_429;
  apr_int64_t e2big_timeout;
} hostinglimits_module_cfg;

// For lost lve enter checking
static __thread int tls_lve_entered = 0;
static __thread int tls_lve_id = 0;
static __thread uint32_t tls_lve_cookie = 0;

static int lve_available = 0;
static int errno_global = 0;
struct liblve *lve = NULL;
static int found_apr_func_ver = 0;
static int use_group = 0;
static int skip_memory = 0;
static int pw_buff = 0;

#if __GNUC__ && (__amd64__ || __amd64 || __x86_64__ || __x86_64) && APR_HAS_SHARED_MEMORY
#define HOSTING_ENTRY_LIMITS 1
#endif

typedef struct {
    volatile apr_time_t limit __attribute__((aligned(16)));
    volatile apr_uint64_t miss_cnt __attribute__((aligned(16)));
    volatile apr_uint64_t hit_cnt __attribute__((aligned(16)));
} hostinglimits_module_entry_limit;

static hostinglimits_module_entry_limit *entry_limits = NULL;
static int entry_limits_enabled = 0;
static apr_shm_t *entry_limits_shm = NULL;
static const char *entry_limits_shm_filename;
static apr_size_t entry_limits_shm_size = 0;

//liblve

APR_DECLARE_OPTIONAL_FN (int, lve_enter_flags,
			 (struct liblve * lve, uint32_t lve_id,
			  uint32_t * cookie, int flags));
APR_OPTIONAL_FN_TYPE (lve_enter_flags) * lve_enter_flags_fn = NULL;
APR_DECLARE_OPTIONAL_FN (int, destroy_lve, (struct liblve * lve));
APR_OPTIONAL_FN_TYPE (destroy_lve) * destroy_lve_fn = NULL;
APR_DECLARE_OPTIONAL_FN (int, lve_exit,
			 (struct liblve * lve, uint32_t * cookie));
APR_OPTIONAL_FN_TYPE (lve_exit) * lve_exit_fn = NULL;
APR_DECLARE_OPTIONAL_FN (struct liblve *, init_lve,
			 (liblve_alloc alloc, liblve_free free));
APR_OPTIONAL_FN_TYPE (init_lve) * init_lve_fn = NULL;
APR_DECLARE_OPTIONAL_FN (int, lve_setup_enter,
			 (struct liblve * lve, uint32_t ve_id,
			  struct liblve_settings * set, uint32_t * cookie,
			  enum liblve_enter_flags flags));
APR_OPTIONAL_FN_TYPE (lve_setup_enter) * lve_setup_enter_fn = NULL;

//apr+lve version 1
APR_DECLARE_OPTIONAL_FN (apr_status_t, apr_lve_environment_init,
			 (int lve_no_maxenter_value, void *lve_ptr,
			  int (*lve_enter_flags_function_ptr) (void *,...),
			  int (*lve_leave_function_ptr) (void *,...),
			  char *suexec_string));
APR_OPTIONAL_FN_TYPE(apr_lve_environment_init) * apr_lve_environment_init_fn = NULL;
//apr+lve version 2
APR_DECLARE_OPTIONAL_FN (apr_status_t, apr_lve_environment_init_group,
			 (int lve_no_maxenter_value, void *lve_ptr,
			  int (*lve_enter_flags_function_ptr) (void *,...),
			  int (*lve_leave_function_ptr) (void *,...),
			  char *suexec_string, int use_group));
APR_OPTIONAL_FN_TYPE(apr_lve_environment_init_group) * apr_lve_environment_init_group_fn = NULL;
//apr+lve version 2
APR_DECLARE_OPTIONAL_FN (apr_status_t, apr_lve_environment_init_group_minuid,
			 (int lve_no_maxenter_value, void *lve_ptr,
			  int (*lve_enter_flags_function_ptr) (void *,...),
			  int (*lve_leave_function_ptr) (void *,...),
			  char *suexec_string, int use_group, int min_uid));
APR_OPTIONAL_FN_TYPE (apr_lve_environment_init_group_minuid) * apr_lve_environment_init_group_minuid_fn = NULL;

/* Paralles */
static int io_limit = 100;
static int cpu_limit = -1;
static int mem_limit = -1;
static int ep_limit = -1;
static int ncpu_limit = 1;

/* Parallels */
typedef struct __attribute__ ((__packed__)) limits_t {
    enum _fields {
        io_lim = 1L << 0
    } fields;

    unsigned int uid;
    unsigned int gid;
    unsigned int lve_cpu;
    unsigned int lve_nproc;
    unsigned int lve_io;
} limits_t;

APR_DECLARE_OPTIONAL_FN(int, poa_cgid_limits_lookup_ex, (conn_rec *, const char *, struct limits_t *));
static APR_OPTIONAL_FN_TYPE(poa_cgid_limits_lookup_ex) * cgid_pfn_limits_lookup_ex = NULL;
/* Parallels */

static int min_uid_cfg = 0;
#define MIN_UID_FILE "/etc/cagefs/cagefs.min.uid"

static void read_cagefs_min_uid() {
    FILE *fp = fopen(MIN_UID_FILE, "rb");
    if (fp) {
        int buffer = 0, rc = 0;
        rc = fread(&buffer, sizeof (int), 1, fp);
        if ((rc > 0) && (buffer > 0)) {
            min_uid_cfg = buffer;
        }
        fclose(fp);
    }
}

static int load_liblve(apr_pool_t * p) {
    void *handle_liblve = dlopen("liblve.so.0", RTLD_LOCAL | RTLD_LAZY);
    if (handle_liblve) {
        lve_enter_flags_fn = dlsym(handle_liblve, "lve_enter_flags");
        destroy_lve_fn = dlsym(handle_liblve, "destroy_lve");
        lve_exit_fn = dlsym(handle_liblve, "lve_exit");
        init_lve_fn = dlsym(handle_liblve, "init_lve");
        lve_setup_enter_fn = dlsym(handle_liblve, "lve_setup_enter");

        if (lve_exit_fn && destroy_lve_fn && lve_enter_flags_fn && init_lve_fn && lve_setup_enter_fn) {
            apr_pool_cleanup_register(p, handle_liblve, (void *) dlclose, apr_pool_cleanup_null);
            return 0;
        }

        dlclose(handle_liblve);
    }
    return -1;
}

//Errors decoding
static char *getErrorUrl(char *buffer, int error_numb) {
    int len;
    if (error_numb < 0)
        error_numb = -error_numb;

    switch (error_numb) {
        case EPERM:
            strncpy(buffer, " MHL-EPERM - apache not started by root", MAX_ERROR_URL_LEN);
            break;
        case E2BIG:
            strncpy(buffer, " MHL-E2BIG - entry processes limit reached", MAX_ERROR_URL_LEN);
            break;
        case ENOMEM:
            strncpy(buffer, " MHL-ENOMEM - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        case EFAULT:
            strncpy(buffer, " MHL-EFAULT - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        case EINVAL:
            strncpy(buffer, " MHL-EINVAL - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        case EBUSY:
            strncpy(buffer, " MHL-EBUSY - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        case ENOSPC:
            strncpy(buffer, " MHL-ENOSPC - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        case ENOENT:
            strncpy(buffer, " MHL-ENOENT - internal LVE error", MAX_ERROR_URL_LEN);
            break;
        default:
            strncpy(buffer, " MHL-unknown - internal LVE error", MAX_ERROR_URL_LEN);
            break;
    }

    len = strlen(buffer);
    strncpy(buffer + len, ", read more: https://docs.cloudlinux.com/cloudlinux_os_components/#errors", MAX_ERROR_URL_LEN - len);

    return buffer;
}

static int getLogLevelByErrno(int error_numb)
{
    switch (error_numb) {
        case E2BIG: return APLOG_WARNING;
        default:    return APLOG_ERR;
    }
}

int get_need_debug(apr_array_header_t *list_of_sites, request_rec *r) {
    int num_names;
    char **names_ptr;
    if (!r->hostname)
        return 0;
    if (list_of_sites) {
        names_ptr = (char **) list_of_sites->elts;
        num_names = list_of_sites->nelts;
        for (; num_names; ++names_ptr, --num_names) {
            if (names_ptr && (strstr(r->hostname, *names_ptr) || (*names_ptr[0] == '*'))) {
                return 1;
            }
        }
    }

    return 0;
}

int get_need_debug_uids(apr_array_header_t *list_of_uids, uid_t uid) {
    int num_names;
    uid_t *names_ptr;

    if (!uid)
        return 0;

    if (list_of_uids) {
        names_ptr = (uid_t *) list_of_uids->elts;
        num_names = list_of_uids->nelts;
        for (; num_names; ++names_ptr, --num_names) {
            if (uid == *names_ptr) {
                return 1;
            }
        }

    }

    return 0;
}

pid_t gettid(void) {
    return syscall(__NR_gettid);
}

static void *hostinglimits_module_merge_config(apr_pool_t * p, void *BASE, void *ADD) {
    hostinglimits_module_cfg *base = BASE;
    hostinglimits_module_cfg *add = ADD;
    hostinglimits_module_cfg *cfg = NULL;

    if (!(cfg = (hostinglimits_module_cfg *) apr_pcalloc(p, sizeof (hostinglimits_module_cfg)))) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, NULL, PREFIX "out of memory");
        return NULL;
    }

    cfg->skip = (add->skip) ? add->skip : base->skip;
    cfg->secure_links = (add->secure_links) ? add->secure_links : base->secure_links;
    cfg->http_error_code = (add->http_error_code != DEFAULT_HTTP_ERROR_CODE) ? add->http_error_code : base->http_error_code;
    cfg->lve_id = (add->lve_id) ? add->lve_id : base->lve_id;
    cfg->allowed_handlers = (add->allowed_handlers) ? add->allowed_handlers : base->allowed_handlers;
    cfg->denied_handlers = (add->denied_handlers) ? add->denied_handlers : base->denied_handlers;
    cfg->err_doc_found = (add->err_doc_found) ? add->err_doc_found : base->err_doc_found;
    cfg->lve_user = (add->lve_user) ? add->lve_user : base->lve_user;
    cfg->debug_sites = (add->debug_sites) ? add->debug_sites : base->debug_sites;
    cfg->debug_uids = (add->debug_uids) ? add->debug_uids : base->debug_uids;
    cfg->itk_uid = (add->itk_uid) ? add->itk_uid : base->itk_uid;
    cfg->ruid_uid = (add->ruid_uid) ? add->ruid_uid : base->ruid_uid;
    cfg->itk_gid = (add->itk_gid) ? add->itk_gid : base->itk_gid;
    cfg->ruid_gid = (add->ruid_gid) ? add->ruid_gid : base->ruid_gid;
    cfg->retryAfter = (add->retryAfter) ? add->retryAfter : base->retryAfter;
    cfg->path_regex = (add->path_regex) ? add->path_regex : base->path_regex;
    cfg->mode = (add->mode) ? add->mode : base->mode;
    cfg->suphp_uid = (add->suphp_uid) ? add->suphp_uid : base->suphp_uid;
    cfg->suphp_gid = (add->suphp_gid) ? add->suphp_gid : base->suphp_gid;
    cfg->use_429 = (add->use_429) ? add->use_429 : base->use_429;
    cfg->e2big_timeout = (add->e2big_timeout) ? add->e2big_timeout : base->e2big_timeout;
    return cfg;
}

static void *hostinglimits_module_create_dir_config(apr_pool_t * p, char *dirspec) {
    hostinglimits_module_cfg *cfg = NULL;

    if (!(cfg = (hostinglimits_module_cfg *) apr_pcalloc(p, sizeof (hostinglimits_module_cfg)))) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, NULL, PREFIX "out of memory");
        return NULL;
    }

    cfg->skip = 0;
    cfg->secure_links = 0;
    cfg->http_error_code = DEFAULT_HTTP_ERROR_CODE;
    cfg->err_doc_found = 0;
    cfg->lve_id = 0;
    cfg->allowed_handlers = NULL;
    cfg->denied_handlers = NULL;
    cfg->lve_user = NULL;
    cfg->debug_sites = NULL;
    cfg->debug_uids = NULL;
    cfg->itk_uid = 0;
    cfg->ruid_uid = 0;
    cfg->itk_gid = 0;
    cfg->ruid_gid = 0;
    cfg->retryAfter = 240;
    cfg->header = NULL;
    cfg->mode = 0;
    cfg->path_regex = apr_pstrdup(p, "/home/([^/]+)/");
    cfg->suphp_uid = 0;
    cfg->suphp_gid = 0;
    cfg->use_429 = 0;
    cfg->e2big_timeout = 0;
    return (void *) cfg;
}

static const char *set_lve_id(cmd_parms * cmd, void *mcfg, const char *lve_id) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (lve_id)
        cfg->lve_id = (uint32_t) apr_atoi64(lve_id);

    return NULL;
}

static const char *set_lve_user(cmd_parms * cmd, void *mcfg, const char *lve_user) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (lve_user)
        cfg->lve_user = lve_user;

    return NULL;
}

static const char *set_assign_user_id(cmd_parms * cmd, void *mcfg, const char *user_name, const char *group_name) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (user_name)
        cfg->itk_uid = ap_uname2id(user_name);

    if (group_name)
        cfg->itk_gid = ap_gname2id(group_name);

    return NULL;
}

static const char *set_uidgid(cmd_parms * cmd, void *mcfg, const char *user_name, const char *group_name) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (user_name)
        cfg->ruid_uid = ap_uname2id(user_name);

    if (group_name)
        cfg->ruid_gid = ap_gname2id(group_name);

    return NULL;
}

static const char *set_debug(cmd_parms * cmd, void *mcfg, const char *arg) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);
    if (err != NULL)
        return err;

    if (!cfg->debug_sites)
        cfg->debug_sites = apr_array_make(cmd->pool, 2, sizeof (char *));

    *(const char **) apr_array_push(cfg->debug_sites) = arg;
    return NULL;
}

static const char *set_debug_uids(cmd_parms * cmd, void *mcfg, const char *arg) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    uid_t uid_arg = apr_atoi64(arg);
    const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);

    if (err != NULL)
        return err;

    if (!cfg->debug_uids)
        cfg->debug_uids = apr_array_make(cmd->pool, 2, sizeof (uid_t));

    if (uid_arg)
        *(uid_t *) apr_array_push(cfg->debug_uids) = uid_arg;

    return NULL;
}

static const char *set_e2big_timeout(cmd_parms * cmd, void *mcfg, const char *arg) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE | NOT_IN_LIMIT);

    if (err != NULL)
        return err;

    if (arg)
        cfg->e2big_timeout = apr_atoi64(arg);
    
    if (cfg->e2big_timeout)
        entry_limits_enabled = 1;

    return NULL;
}


#ifndef APACHE2_0
// Function handles Apache's ErrorDocument directive
// Sets err_doc_found flag in module configuration if "ErrorDocument 508" directive is found

static const char *error_doc_func(cmd_parms * cmd, void *mcfg, const char *par1, const char *par2) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (par1 && strstr(par1, "508"))
        cfg->err_doc_found = 1;

    return NULL;
}
#endif

static unsigned int get_valid_http_error_code(unsigned int code) {
    if ((code < MIN_HTTP_SERVER_ERROR_CODE) || (code > MAX_HTTP_SERVER_ERROR_CODE))
        return DEFAULT_HTTP_ERROR_CODE;

    return code;
}

// Function handles LVEErrorCode directive
static const char *set_lve_error_code(cmd_parms * cmd, void *mcfg, const char *lve_error_code) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (lve_error_code)
        cfg->http_error_code = get_valid_http_error_code((unsigned int) apr_atoi64(lve_error_code));

    return NULL;
}

static const char *set_lve_retryafter(cmd_parms * cmd, void *mcfg, const char *minutes) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (minutes)
        cfg->retryAfter = (uint32_t) apr_atoi64(minutes);

    return NULL;
}

static const char *set_lve_parsemode(cmd_parms * cmd, void *mcfg, const char *mode) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (mode) {
        if (!apr_strnatcasecmp(mode, "CONF")) {
            cfg->mode = 0;
            return NULL;
        }
        if (!apr_strnatcasecmp(mode, "PATH")) {
            cfg->mode = 1;
            return NULL;
        }
        if (!apr_strnatcasecmp(mode, "OWNER")) {
            cfg->mode = 2;
            return NULL;
        }
        /*  
        if (!apr_strnatcasecmp (mode, "HEADER")) {
            cfg->mode = 3;
            return NULL;
        } 
         */
    }

    return NULL;
}

static const char *set_lve_pathregexp(cmd_parms * cmd, void *mcfg, const char *regexp_data) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (regexp_data) {
        int right_param = 1;
        ap_regex_t rx;

        // Compile regex. Error ?
        if (ap_regcomp(&rx, regexp_data, AP_REG_EXTENDED)) {
            right_param = 0;
        } else {
            cfg->path_regex = regexp_data;
            ap_regfree(&rx);
        }

        if (!right_param)
            return apr_psprintf(cmd->pool, "Wrong regexp expression %s in parameter LVEPathRegexp", regexp_data);
    }
    return NULL;
}

static const char *set_lve_headername(cmd_parms * cmd, void *mcfg, const char *header) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    if (header)
        cfg->header = (char *) header;

    return NULL;
}

static const char *set_handlers(cmd_parms * cmd, void *mcfg, const char *arg) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    const char *err = ap_check_cmd_context(cmd, NOT_IN_LIMIT);
    if (err != NULL)
        return err;

    if (!cfg->allowed_handlers)
        cfg->allowed_handlers = apr_array_make(cmd->pool, 2, sizeof (char *));

    *(const char **) apr_array_push(cfg->allowed_handlers) = arg;
    return NULL;
}

static const char *set_handlers_to_deny(cmd_parms * cmd, void *mcfg, const char *arg) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
    const char *err = ap_check_cmd_context(cmd, NOT_IN_LIMIT);
    if (err != NULL)
        return err;

    if (!cfg->denied_handlers)
        cfg->denied_handlers = apr_array_make(cmd->pool, 2, sizeof (char *));

    *(const char **) apr_array_push(cfg->denied_handlers) = arg;
    return NULL;
}

#ifndef APACHE2_0
static const char *suphp_handle_cmd_user_group_lve(cmd_parms * cmd, void *mconfig, const char *arg1, const char *arg2) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mconfig;
    if (arg1)
        cfg->suphp_uid = ap_uname2id(arg1);

    if (arg2)
        cfg->suphp_gid = ap_gname2id(arg2);

    return NULL;
}
#endif

static const char *set_lve_use_group(cmd_parms * cmd, void *mcfg, const char *arg) {
    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
    if (err != NULL)
        return err;

    if (!apr_strnatcasecmp(arg, "On"))
        use_group = 1;

    return NULL;
}

static const char *set_lve_skip_memory(cmd_parms * cmd, void *mcfg, const char *arg) {
    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
    if (err != NULL)
        return err;

    if (!apr_strnatcasecmp(arg, "On"))
        skip_memory = 1;

    return NULL;
}

/* Parallels */
static const char *set_cpu_limit(cmd_parms * cmd, void *dummy, const char *arg) {
    if (arg)
        cpu_limit = atol(arg);

    return NULL;
}

static const char *set_ncpu_limit(cmd_parms * cmd, void *dummy, const char *arg) {
    if (arg)
        ncpu_limit = atol(arg);

    return NULL;
}

static const char *set_mem_limit(cmd_parms * cmd, void *dummy, const char *arg) {
    if (arg)
        mem_limit = atol(arg);

    return NULL;
}

static const char *set_ep_limit(cmd_parms * cmd, void *dummy, const char *arg) {
    if (arg)
        ep_limit = atol(arg);

    return NULL;
}

static const char *set_io_limit(cmd_parms * cmd, void *dummy, const char *arg) {
    if (arg)
        io_limit = atol(arg);

    return NULL;
}

/* Parallels */

//Deprecated
static const char *
set_lve_use_filter (cmd_parms * cmd, void *mcfg, const char *arg)
{
  return NULL;
}

// describe used directives
static command_rec hostinglimits_module_directives[] = {
    AP_INIT_TAKE1("LVELimitRecheckTimeout", set_e2big_timeout, NULL, RSRC_CONF, "Timeout before recheck LVE after resource limit reached"),
    AP_INIT_FLAG("SkipErrors", ap_set_flag_slot, (void *) APR_OFFSETOF(hostinglimits_module_cfg, skip), 
            RSRC_CONF, "Allow apache to continue even if LVE is unavalable"),
    AP_INIT_FLAG("LVEUse429", ap_set_flag_slot, (void *) APR_OFFSETOF(hostinglimits_module_cfg, use_429), 
            ACCESS_CONF | RSRC_CONF, "Use 429 error code instead 508"),
    AP_INIT_TAKE1("LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, "LVE Id"),
    AP_INIT_TAKE1("LVEUser", set_lve_user, NULL, ACCESS_CONF | RSRC_CONF, "LVE User"),
    AP_INIT_ITERATE("AllowedHandlers", set_handlers, NULL, ACCESS_CONF | RSRC_CONF, "A list of handlers which will be put in LVE"),
    AP_INIT_ITERATE("DenyHandlers", set_handlers_to_deny, NULL, ACCESS_CONF | RSRC_CONF, "A list of handlers which will be NOT put in LVE"),
    AP_INIT_TAKE1("LVEErrorCode", set_lve_error_code, NULL, RSRC_CONF, "Integer HTTP error code to return to client if resources limit is reached"),
#ifndef APACHE2_0
    AP_INIT_TAKE2("ErrorDocument", error_doc_func, NULL, OR_FILEINFO, "ErrorDocument Apache configuration directive"),
    AP_INIT_FLAG("SecureLinks", ap_set_flag_slot, (void *) APR_OFFSETOF(hostinglimits_module_cfg, secure_links), RSRC_CONF, "Make sure that files and links owned by same user as defined for VirtualHost"),
    AP_INIT_TAKE2("suPHP_UserGroup", suphp_handle_cmd_user_group_lve, NULL, RSRC_CONF | ACCESS_CONF, "User and group scripts shall be run as"),
#endif
    AP_INIT_TAKE2("AssignUserID", set_assign_user_id, NULL, RSRC_CONF | ACCESS_CONF, "Tie a virtual host to a specific child process."),
    AP_INIT_TAKE2("RUidGid", set_uidgid, NULL, RSRC_CONF | ACCESS_CONF, "Minimal uid or gid file/dir, else set[ug]id to default (User,Group)"),
    AP_INIT_ITERATE("LVESitesDebug", set_debug, NULL, RSRC_CONF, "A list of sites which should be debugged"),
    AP_INIT_ITERATE("LVEUidsDebug", set_debug_uids, NULL, RSRC_CONF, "A list of uids which should be debugged"),
    AP_INIT_TAKE1("LVERetryAfter", set_lve_retryafter, NULL, RSRC_CONF, "Set minutes of LVE_RETRY_AFTER header, which comes with 508 error. O - disabled"),
    AP_INIT_TAKE1("LVEParseMode", set_lve_parsemode, NULL, RSRC_CONF, "Set mode of uid extraction - CONF|PATH|OWNER|HEADER"),
    AP_INIT_TAKE1("LVEPathRegexp", set_lve_pathregexp, NULL, RSRC_CONF, "Set template of path for mode PATH"),
    AP_INIT_TAKE1("LVEHeaderName", set_lve_headername, NULL, RSRC_CONF, "Set header name for uid extraction for mode Header"),
    AP_INIT_TAKE1("LVEUseGroupID", set_lve_use_group, NULL, RSRC_CONF, "Use group ID instead of UID for LVE id"),
    AP_INIT_TAKE1("LVESkipMemory", set_lve_skip_memory, NULL, RSRC_CONF, "Starts using LVE_NO_UBC flag to enter LVE which disables io and mem limit"),

    /* Paralles */
    AP_INIT_TAKE1("SuCGIDIOLimit", set_io_limit, NULL, RSRC_CONF, "IO Limit in KB/s"),
    AP_INIT_TAKE1("OverrideCPULimit", set_cpu_limit, NULL, RSRC_CONF, "CPU Limit"),
    AP_INIT_TAKE1("OverrideMEMLimit", set_mem_limit, NULL, RSRC_CONF, "MEM Limit"),
    AP_INIT_TAKE1("OverrideEPROCLimit", set_ep_limit, NULL, RSRC_CONF, "EP Limit"),
    AP_INIT_TAKE1("OverrideNCPULimit", set_ncpu_limit, NULL, RSRC_CONF, "NCPU Limit"),
    AP_INIT_TAKE1 ("LVEUseFilter", set_lve_use_filter, NULL,
		 RSRC_CONF,
		 "Use filter for LVE out(deprecated)"),
    /* Paralles */
    {NULL}
};

static apr_status_t cleanup_entry_limits_shm(void *not_used) {
    ap_log_error(APLOG_MARK, APLOG_INFO, 0, NULL, PREFIX "cleaning up shared memory");

    if (entry_limits_shm) {
        apr_shm_destroy(entry_limits_shm);
        entry_limits_shm = NULL;
        entry_limits = NULL;
    }

    return APR_SUCCESS;
}

#if !AP_MODULE_MAGIC_AT_LEAST(20120211, 2)
static char *ap_runtime_dir_relative(apr_pool_t *p, const char *file) {
    char *newpath = NULL;
    apr_status_t rv;
    const char *runtime_dir = ap_server_root_relative(p, DEFAULT_REL_RUNTIMEDIR);

    rv = apr_filepath_merge(&newpath, runtime_dir, file,
            APR_FILEPATH_TRUENAME, p);
    if (newpath && (rv == APR_SUCCESS || APR_STATUS_IS_EPATHWILD(rv)
            || APR_STATUS_IS_ENOENT(rv)
            || APR_STATUS_IS_ENOTDIR(rv))) {
        return newpath;
    } else {
        return NULL;
    }
}
#endif

static int hostinglimits_post_config(apr_pool_t * pconf, apr_pool_t * plog, apr_pool_t * ptemp, server_rec * s) {
    apr_status_t sts;
    apr_uintptr_t entry_limit_idx = 0;

    /*
     * check if need force lve exit on filter
     */

    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "use Min UID %d", min_uid_cfg);

    /* Parallels */
    cgid_pfn_limits_lookup_ex = APR_RETRIEVE_OPTIONAL_FN(poa_cgid_limits_lookup_ex);
    /* Parallels */

    if (lve_available) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "version " MOD_HOSTINGLIMITS_VERSION ". LVE mechanism enabled");
        if (use_group)
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "use GroupID instead of UID");

        if (found_apr_func_ver) {
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "found apr extention version %d", found_apr_func_ver);
            if (apr_lve_environment_init_group_minuid_fn) {
                apr_lve_environment_init_group_minuid_fn(/* LVE_NO_MAXENTER|LVE_SILENCE */ (int) ((1 << 2) | (1 << 3)),
                        (void *) lve, (void *) lve_enter_flags_fn, (void *) lve_exit_fn, SUEXEC_BIN, use_group, min_uid_cfg);
                ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "apr_lve_environment_init_group_minuid check ok");
            } else if (apr_lve_environment_init_group_fn) {
                apr_lve_environment_init_group_fn(/* LVE_NO_MAXENTER|LVE_SILENCE */ (int) ((1 << 2) | (1 << 3)), 
                        (void *) lve, (void *) lve_enter_flags_fn, (void *) lve_exit_fn, SUEXEC_BIN, use_group);
                ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "apr_lve_environment_init_group check ok");
            } else if (apr_lve_environment_init_fn) {
                apr_lve_environment_init_fn(/* LVE_NO_MAXENTER|LVE_SILENCE */ (int) ((1 << 2) | (1 << 3)), 
                        (void *) lve, (void *) lve_enter_flags_fn, (void *) lve_exit_fn, SUEXEC_BIN);
                ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s, PREFIX "apr_lve_environment_init check ok");
            }
        } else {
            ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s, PREFIX "apr_lve_* not found! Please, contact support at http://www.cloudlinux.com/support/");
        }

    } else {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, PREFIX "version " MOD_HOSTINGLIMITS_VERSION ". LVE mechanism disabled, LVE is unavailable. Errno %d", errno_global);
    }

#ifdef HOSTING_ENTRY_LIMITS
    if (entry_limits || !entry_limits_enabled) return OK;

    while (s) {
        entry_limit_idx++;
        ap_set_module_config(s->module_config, &hostinglimits_module, (void *)entry_limit_idx);
        s = s->next;
    }

    if (entry_limits_shm_size = sizeof (hostinglimits_module_entry_limit) * entry_limit_idx) {
        hostinglimits_module_entry_limit *e;

        /*
         * Create a unique filename using our pid. This information is
         * stashed in the global variable so the children inherit it.
         */

        entry_limits_shm_filename = ap_runtime_dir_relative(pconf, "hostinglimits_shm");
        entry_limits_shm_filename = ap_append_pid(pconf, entry_limits_shm_filename, ".");

        /* Now create that segment */
        sts = apr_shm_create(&entry_limits_shm, entry_limits_shm_size, entry_limits_shm_filename, pconf);
        if (APR_SUCCESS != sts) {
            ap_log_error(APLOG_MARK, APLOG_ERR, sts, s, PREFIX
                    "Failed to create shared memory segment on file %s",
                    entry_limits_shm_filename);
            return !OK;
        }

        e = entry_limits = (hostinglimits_module_entry_limit *)apr_shm_baseaddr_get(entry_limits_shm);
        while(entry_limit_idx) {
            e->limit = 0;
            e++;
            entry_limit_idx--;
        }

        /* Call cleanup_entry_limits_shm on exit or restart */
        apr_pool_cleanup_register(pconf, NULL, cleanup_entry_limits_shm, apr_pool_cleanup_null);
    }
#endif  /* APR_HAS_SHARED_MEMORY */

    return OK;
}

#define MAX_REGEX_LEN 255

// Function searches handler in handler_list
// returns 1 if found, 0 if not

static int match_handler(apr_array_header_t * handlers_list, const char *handler) {
    int num_names;
    char **names_ptr;
    char *regex_begin_ptr; // Pointer to "%" character in the beginning of regex
    char *regex_end_ptr; // Pointer to "%" character in the end of regex
    char regex_str[MAX_REGEX_LEN + 1]; // Buffer to copy regex from handler_list
    int regex_len; // Length of regex

    ap_regex_t compiled_regex; // Compiled regex buffer

    if (handlers_list && handler) {
        names_ptr = (char **) handlers_list->elts;
        num_names = handlers_list->nelts;

        // Scan handlers_list
        for (; num_names; ++names_ptr, --num_names) { // Match all the handlers ?
            if (!strcmp("*", *names_ptr))
                return 1;

            // Current string in handlers_list is regex ?
            if (regex_begin_ptr = strchr(*names_ptr, '%')) {
                // Get pointer to "%" character in the end of regex
                regex_end_ptr = strchr(*names_ptr + 1, '%');

                // End of regex is not found ?
                if (!regex_end_ptr)
                    continue;

                // Calculate regex length
                regex_len = regex_end_ptr - regex_begin_ptr - 1;

                // Regex is too short or too long ?
                if ((regex_len < 1) || (regex_len > MAX_REGEX_LEN))
                    continue;

                // Make copy of regex
                strncpy(regex_str, regex_begin_ptr + 1, regex_len);
                regex_str[regex_len] = '\0';

                // Compile regex. Error ?
                if (ap_regcomp(&compiled_regex, regex_str, AP_REG_EXTENDED | AP_REG_NOSUB))
                    continue;

                // Match handler against compiled regex. Match is found ?
                if (!ap_regexec(&compiled_regex, handler, 0, NULL, 0)) {
                    ap_regfree(&compiled_regex);
                    return 1;
                } else {
                    ap_regfree(&compiled_regex);
                    continue;
                }
            }

            // Compare strings literally
            if (!strcmp(handler, *names_ptr))
                return 1;
        }
    }

    return 0;
}

#ifdef APACHE2_0
/**
 * Find DocumentError from mod_core
 */
int readErrorDocument(request_rec * r, hostinglimits_module_cfg * cfg) {
    core_dir_config *conf = ap_get_module_config(r->per_dir_config, &core_module);
    if (conf) {
        if (conf->response_code_strings) {
            int index_number, idx500 =
                    ap_index_of_response(HTTP_INTERNAL_SERVER_ERROR);
            if ((index_number =
                    ap_index_of_response(DEFAULT_HTTP_ERROR_CODE)) == idx500) {
                return 0;
            }
            if (!conf->response_code_strings[index_number]) {
                return 0;
            }
            return 1;
        }
    }
    return 0;
}
#endif

static int process_lve_error(request_rec * r, hostinglimits_module_cfg * cfg) {

    //Set out header LVE_RETRY_AFTER
    if (cfg->retryAfter)
        apr_table_setn(r->err_headers_out, "Retry-After", apr_ltoa(r->pool, cfg->retryAfter * 60));

#ifdef APACHE2_4
    if (cfg->use_429 && cfg->http_error_code == DEFAULT_HTTP_ERROR_CODE)
        return DEFAULT_HTTP_ERROR_CODE_NEW;
#endif

#ifdef APACHE2_0
    // LVEErrorCode is NOT 508 or "ErrorDocument 508" directive is present in Apache configuration ?
    if ((cfg->http_error_code != DEFAULT_HTTP_ERROR_CODE) || (readErrorDocument(r, cfg)))
        return cfg->http_error_code;
#else
    // LVEErrorCode is NOT 508 or "ErrorDocument 508" directive is present in Apache configuration ?
    if ((cfg->http_error_code != DEFAULT_HTTP_ERROR_CODE) || (cfg->err_doc_found))
        return cfg->http_error_code;
#endif


#ifndef APACHE2_4
    r->status = cfg->use_429?DEFAULT_HTTP_ERROR_CODE_NEW:DEFAULT_HTTP_ERROR_CODE;
#else
    r->status = DEFAULT_HTTP_ERROR_CODE;
#endif
    r->content_type = "text/html";
    apr_bucket_brigade *bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
    ap_basic_http_header(r, bb);

    // Send header information only (HEAD request) ?
    if (r->header_only)
        return DONE;

#ifndef APACHE2_4
    ap_rvputs(r, "\n" DOCTYPE_HTML_2_0
            "<HTML><HEAD>\n<TITLE>",  (cfg->use_429?" 429 Too Many Requests</TITLE>\n":" 508 Resource Limit Is Reached</TITLE>\n"),
            "</HEAD><BODY>\n", (cfg->use_429?"<H1>Too Many Requests</H1>\n":"<H1>Resource Limit Is Reached</H1>\n"), NULL);
    if (cfg->use_429)
        ap_rputs("The website is temporarily unable to service your request as it exceeded requests limit.\n"
            "Please try again later.\n", r);
    else
        ap_rputs("The website is temporarily unable to service your request as it exceeded resource limit.\n"
            "Please try again later.\n", r);
#else
    ap_rvputs(r, "\n" DOCTYPE_HTML_2_0
            "<HTML><HEAD>\n<TITLE>508 Resource Limit Is Reached</TITLE>\n"
            "</HEAD><BODY>\n" "<H1>Resource Limit Is Reached</H1>\n", NULL);

    ap_rputs("The website is temporarily unable to service your request as it exceeded resource limit.\n"
            "Please try again later.\n", r);
#endif
    ap_rputs(ap_psignature("<HR>\n", r), r);
    ap_rputs("</BODY></HTML>\n", r);
    ap_finalize_request_protocol(r);
    ap_rflush(r);
    return DONE;
}

char *get_regexp_match(apr_pool_t * pool, ap_regex_t * rx, char *buf, int match) {
    int result;
    ap_regmatch_t *matches;
    char *bbuf;
    if (rx->re_nsub < match) {
        return NULL;
    }
    matches = (ap_regmatch_t *) apr_palloc(pool, (rx->re_nsub + 1) * sizeof (ap_regmatch_t));
    if (!matches)
        return NULL;

    if (!buf || !buf[0])
        return NULL;
    result = ap_regexec(rx, buf, rx->re_nsub + 1, matches, 0);
    if (!result) {
        int i;
        for (i = 0; i <= rx->re_nsub; i++) {
            if ((matches[i].rm_so != -1) && (i == match)) {
                bbuf =
                        apr_psprintf(pool, "%.*s",
                        matches[i].rm_eo - matches[i].rm_so,
                        buf + matches[i].rm_so);
                return bbuf;
            }
        }
    }
    return NULL;
}

int getEnvironmentValue(request_rec * r, hostinglimits_module_cfg * cfg) {
    const char *result = NULL;
    if (r->subprocess_env) {
        if (cfg->header)
            result = apr_table_get(r->subprocess_env, (const char *) cfg->header);
        else
            result = apr_table_get(r->subprocess_env, (const char *) X_LVE_ID);
    }
    if (result)
        return (int) ap_uname2id(result);

    return 0;
}

#ifdef APACHE2_0
typedef struct {
    int engine; // Status of suPHP_Engine
    char *php_config;
    int cmode; // Server of directory configuration?
    char *target_user;
    char *target_group;
    apr_table_t *handlers;
    char *php_path;
} suphp_conf_lve;
#endif

uid_t get_suphp_uid(request_rec * r, hostinglimits_module_cfg * cfg, int use_grp) {
#ifndef APACHE2_0
    if (use_grp)
        return (uid_t) cfg->suphp_gid;
    else
        return cfg->suphp_uid;
#else
    module *suphp_c = NULL;
    suphp_c = ap_find_linked_module(SUPHP_MODULE);
    if (suphp_c) {
        suphp_conf_lve *pconf = (suphp_conf_lve *) ap_get_module_config(r->per_dir_config, suphp_c);
        if (pconf && pconf->target_user && pconf->target_group) {
            if (use_grp)
                return ap_gname2id(pconf->target_group);
            else
                return ap_uname2id(pconf->target_user);
        }
    }
    return 0;
#endif
}

/* Use group ID
 * LVEId/LVEUser + UseGroupID - not work (always standard mode)
 * Suexec, Ruid2, itk, SuPHP + UseGroupID - work (UID/GID)
 * PATH + UseGroupID - not work (always standard mode)
 * OWNER + UseGroupID - work (user or group of file)
 * HEADER + UseGroupID - not work (always standard mode)
 */
int get_lve_id(request_rec * r, hostinglimits_module_cfg * cfg) {
    int lve_id;
    uid_t uid = 0;
    //Priority steps
    // 1) LVEId
    // 2) LVEUser

    //enter into lve
    //1) check if set LVEId
    lve_id = cfg->lve_id;
    if (lve_id > 0)
        return lve_id;
    //2) if LVEId is unset and set LVEUser, get id here

    if (cfg->lve_user) {
        //if LVEUser set on, read it
        lve_id = (int) ap_uname2id(cfg->lve_user);
        if (lve_id < 1)
            return 0;
        return lve_id;
    }

    switch (cfg->mode) {
        case 0:
        {
            //Priority steps
            // 3) SuexecUserGroup
            // 4) RUidGid
            // 5) AssignUserID
            // 6) suPHP_UserGroup

            //Get uid from suexec
            ap_unix_identity_t *ugid = ap_run_get_suexec_identity(r);
            if (ugid != NULL)
                if (use_group)
                    return ugid->gid;
                else
                    return ugid->uid;

            //4) still less then need, than check ruid2
            //check for ruid2
            if (cfg->ruid_uid > 0)
                if (use_group)
                    return cfg->ruid_gid;
                else
                    return cfg->ruid_uid;

            //5) still less then need, than check itk
            //check for itk, if lve_id still less then 0
            if (cfg->itk_uid > 0)
                if (use_group)
                    return cfg->itk_gid;
                else
                    return cfg->itk_uid;

            //6) still less then need, than check suPHP
            uid_t tmp_suphp = get_suphp_uid(r, cfg, use_group);
            if (tmp_suphp > 0)
                return tmp_suphp;
        }
            return 0;

        case 1: //PATH
        {
            char *user_name = NULL;
            //Parse path
            apr_finfo_t st;
            //Check if file exists
            if (apr_stat(&st, r->filename, APR_FINFO_NORM, r->pool) ==
                    APR_SUCCESS) {
                int need_prcd = 1;
                ap_regex_t rx;
                if (ap_regcomp(&rx, cfg->path_regex, AP_REG_EXTENDED)) {
                    need_prcd = 0;
                }
                //Get user name from path
                if (need_prcd) {
                    user_name = get_regexp_match(r->pool, &rx, r->filename, 1);
                    ap_regfree(&rx);
                }
                int find_uid = 0;
                if (user_name) {
                    //Find user id
                    find_uid = (int) ap_uname2id(user_name);
                    if (find_uid > 0)
                        return find_uid;
                }
            }
        }
            return 0;
        case 2: //OWNER
        {
            apr_finfo_t st;
            if (apr_stat(&st, r->filename, APR_FINFO_NORM, r->pool) ==
                    APR_SUCCESS) {
                if (use_group)
                    return (int) st.group;
                else
                    return (int) st.user;
            }
        }
            return 0;
        case 3: //HEADER
        {
            /*
             * There are exists to ways
             * 1) one way - to read environment variables, safeless way
             * because variable can be set into .htaccess
             * 2) on the beginning of the process - to read mod_env
             * configuration into array per vhost. We lost per-dir configuration
             * in this case
             */

            return getEnvironmentValue(r, cfg);

        }
            break;
    }
    return -1;
}

#ifndef strnlen
size_t strnlen(const char *s, size_t maxlen) {
    size_t len;
    for (len = 0; len < maxlen; len++, s++) {
        if (!*s)
            break;
    }
    return (len);
}
#endif

static int mod_hostinglimits_file_exist(char *filename) {
    struct stat buffer;
    if (filename)
        return (stat(filename, &buffer) == 0);

    return 0;
}

typedef struct {
    int need_debug;
    int suspended;
    int lve_entered;
    uint32_t lve_cookie;
    int lve_id;
    int is_paralles;
    enum liblve_enter_flags p_flags;
    struct liblve_settings lve_data;
} hostinglimits_request_cfg;


int hostinglimits_create_request(request_rec *r) {
    hostinglimits_request_cfg *rc = NULL;

    if (r->main) {
        rc = (hostinglimits_request_cfg *)ap_get_module_config(r->main->request_config, &hostinglimits_module);
    } else if (r->prev) {
        rc = (hostinglimits_request_cfg *)ap_get_module_config(r->prev->request_config, &hostinglimits_module);
    } else {
        rc = (hostinglimits_request_cfg *)apr_pcalloc(r->pool, sizeof(hostinglimits_request_cfg));
    }

    if (rc)
        ap_set_module_config(r->request_config, &hostinglimits_module, rc);

    return OK;
}

/**
 * Where are main functionality in this function.
 * This funcion performs after reading request headers
 * and before other processing.
 *
 *
 */
static int hostinglimits_handler(request_rec * r) {
    hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config, &hostinglimits_module);
    hostinglimits_request_cfg *rcfg = (hostinglimits_request_cfg *)ap_get_module_config(r->request_config, &hostinglimits_module);
    hostinglimits_module_entry_limit *entry_limit = NULL;
    apr_time_t elimit = 0, ct = 0, e2big_timeout = cfg->e2big_timeout ? cfg->e2big_timeout * 1000 : 0;
    apr_uint32_t elimit_hit = 0;
    char err_message_url[MAX_ERROR_URL_LEN];
    struct limits_t limits_data;
    int rc = 0, lve_rc = 0;
    int keep_errno = 0;

    if (!lve_available)
        return DECLINED;

    if (!cfg) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX
                "No request per dir config. Posible out of memory: HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) min_uid (%d)",
                r->handler, r->hostname, r->uri, gettid(), min_uid_cfg);

        return HTTP_INTERNAL_SERVER_ERROR;
    }

    if (!rcfg) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX
                "No request config. Posible out of memory: HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) min_uid (%d)",
                r->handler, r->hostname, r->uri, gettid(), min_uid_cfg);

        return HTTP_INTERNAL_SERVER_ERROR;
    }

    if (rcfg->lve_entered) {
        if (rcfg->need_debug) {
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                    "[DEBUG] STEP#0001.1 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                    rcfg->lve_id, rcfg->lve_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool ? r->pool : 0);
        }

        return DECLINED;
    }

    rcfg->is_paralles = 0;

    /* Paralles */
    if (cgid_pfn_limits_lookup_ex && !cgid_pfn_limits_lookup_ex(r->connection, r->hostname, &limits_data)) {
        core_dir_config *CORE_conf = ap_get_module_config(r->per_dir_config, &core_module);
        rcfg->lve_id = limits_data.uid;
        // rcfg->lve_data = {-1, -1, -1, -1, -1, -1, -1, -1};

        rcfg->lve_data.ls_cpu = (cpu_limit == -1) ? limits_data.lve_cpu : cpu_limit;
        rcfg->lve_data.ls_cpus = (ncpu_limit == -1) ? 0 : ncpu_limit;
        rcfg->lve_data.ls_io = (limits_data.fields & io_lim) ? limits_data.lve_io : io_limit;
        rcfg->lve_data.ls_enters = (ep_limit == -1) ? limits_data.lve_nproc : ep_limit;
        rcfg->lve_data.ls_memory = 0;
        rcfg->lve_data.ls_cpu_weight = 100;
        //256 = (1024 * 1024)/1024/4
        rcfg->lve_data.ls_memory_phy = CORE_conf->limit_mem ? (CORE_conf->limit_mem->rlim_max / 1024 / 4) : ((mem_limit != -1) ? (mem_limit * 256) : -1);
        rcfg->lve_data.ls_nproc = (ep_limit == -1 ? limits_data.lve_nproc : ep_limit);
        rcfg->lve_data.ls_iops = 0;

        if (cfg->lve_id > 0)
            rcfg->lve_id = cfg->lve_id;

        if ((ep_limit == -1 ? limits_data.lve_nproc : ep_limit) <= 0)
            rcfg->p_flags |= LVE_NO_MAXENTER;

        rcfg->is_paralles = 1;
    }
    /* Parallels */

    if (!rcfg->is_paralles)
        rcfg->lve_id = get_lve_id(r, cfg);

    if (!rcfg->need_debug)
        rcfg->need_debug = get_need_debug(cfg->debug_sites, r);

    if (!rcfg->need_debug)
        rcfg->need_debug = get_need_debug_uids(cfg->debug_uids, rcfg->lve_id);

    if (rcfg->need_debug) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                "[DEBUG] STEP#0001.2 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                rcfg->lve_id, rcfg->lve_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool ? r->pool : 0);
    }

    if (rcfg->lve_id < min_uid_cfg || !rcfg->lve_id) {
        if (rcfg->need_debug) {
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                    "[DEBUG] STEP#0002.1 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                    rcfg->lve_id, rcfg->lve_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool ? r->pool : 0);
        }

        return DECLINED;
    }

    //lets find handlers for which HostingLimits should NOT run...
    // Check if handler is NOT in allowed_handlers list or if handler is in denied_handlers list
    if ((!match_handler(cfg->allowed_handlers, r->handler)) || match_handler(cfg->denied_handlers, r->handler)) {
        if (rcfg->need_debug) {
            ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                    "[DEBUG] STEP#0002.2 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                    rcfg->lve_id, rcfg->lve_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool ? r->pool : 0);
        }

        return DECLINED;
    }

    char *tmp_res = (char *) apr_table_get(r->notes, X_LVE_ID_HEADER);
    if (!tmp_res || rcfg->lve_id != (int) apr_atoi64(tmp_res))
        apr_table_set(r->notes, X_LVE_ID_HEADER, apr_itoa(r->pool, rcfg->lve_id));

    if (tls_lve_entered) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX
                "Already inside LVE: CURRENT-LVE(%d) LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) min_uid (%d)",
                tls_lve_id, rcfg->lve_id, r->handler, r->hostname, r->uri, gettid(), min_uid_cfg);

        if (lve_rc = (*lve_exit_fn) (lve, &tls_lve_cookie)) {
            ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX "LVE exiting error (%d)", lve_rc);
        }
    }

#ifdef HOSTING_ENTRY_LIMITS
    if (entry_limits && e2big_timeout) {
        apr_uintptr_t idx;
        if (idx = (apr_uintptr_t) ap_get_module_config(r->server->module_config, &hostinglimits_module)) {
            ct = apr_time_now();
            entry_limit = &entry_limits[idx-1];
            elimit = entry_limit->limit;

            if (elimit > ct)
                elimit_hit = 1;
        }
    }
#endif

    if (skip_memory)
        rcfg->p_flags |= LVE_NO_UBC;

    if (entry_limit && elimit_hit) {
        lve_rc = keep_errno = E2BIG;
    } else 
        if (rcfg->is_paralles) {
        lve_rc = (*lve_setup_enter_fn) (lve, rcfg->lve_id, &rcfg->lve_data, &rcfg->lve_cookie, rcfg->p_flags);
        keep_errno = errno;
    } else {
        lve_rc = (*lve_enter_flags_fn) (lve, rcfg->lve_id, &rcfg->lve_cookie, rcfg->p_flags);
        keep_errno = errno;
    }

    if (lve_rc) {
#ifdef HOSTING_ENTRY_LIMITS
        if (entry_limit && keep_errno == E2BIG) {
            if (!elimit_hit) {
                __sync_add_and_fetch(&entry_limit->miss_cnt, 1);

                while (elimit < ct)
                    elimit = __sync_val_compare_and_swap(&entry_limit->limit, elimit, ct + e2big_timeout);
            } else
                __sync_add_and_fetch(&entry_limit->hit_cnt, 1);

            if (rcfg->need_debug) {
                ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                        "[DEBUG] STEP#0004 LVE(%d) HANDLER(%s) HOSTNAME(%s) TID(%d) HIT(%d) MISS(%d)",
                        rcfg->lve_id, r->handler, r->hostname, gettid(), entry_limit->hit_cnt, entry_limit->miss_cnt);
            }
        }
#endif

        if (keep_errno == EPERM) { //if already inside LVE
            ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX
                    "Already inside LVE: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno (%d)%s min_uid (%d)",
                    rcfg->lve_id, r->handler, r->hostname, r->uri,
                    gettid(), keep_errno,
                    getErrorUrl(err_message_url, keep_errno),
                    min_uid_cfg);

            return DECLINED;
        }

        ap_log_error(APLOG_MARK, getLogLevelByErrno(keep_errno), OK, r->server, PREFIX
                "Error on LVE enter: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno (%d)%s min_uid (%d)",
                rcfg->lve_id, r->handler, r->hostname, r->uri, gettid(),
                keep_errno, getErrorUrl(err_message_url, keep_errno),
                min_uid_cfg);

        return process_lve_error(r, cfg);
    }

    tls_lve_id = rcfg->lve_id;
    tls_lve_cookie = rcfg->lve_cookie;
    tls_lve_entered = rcfg->lve_entered = 1;

    rc = ap_run_handler(r);

    if (lve_rc = (*lve_exit_fn) (lve, &rcfg->lve_cookie)) {
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server, PREFIX "Error exiting %d", lve_rc);
    }

    tls_lve_entered = tls_lve_id = tls_lve_cookie = rcfg->lve_entered = 0;

    if (rcfg->need_debug) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server, PREFIX
                "[DEBUG] STEP#0003 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                rcfg->lve_id, rcfg->lve_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool ? r->pool : 0);
    }

    return rc;
}

apr_pool_t *lve_pool;
void *lve_alloc (int size)
{
  return apr_palloc (lve_pool, size);
}

static char *strnstr(const char *s, const char *find, size_t slen) {
    char c, sc;
    size_t len;

    if ((c = *find++) != '\0') {
        len = strlen(find);
        do {
            do {
                if (slen-- < 1 || (sc = *s++) == '\0')
                    return (NULL);
            } while (sc != c);
            if (len > slen)
                return (NULL);
        } while (strncmp(s, find, len) != 0);
        s--;
    }
    return ((char *) s);
}

static int callback(struct dl_phdr_info *info, size_t size, void *data) {
    apr_pool_t *p = (apr_pool_t *) data;

#ifdef APACHE2_2
    if (strnstr(info->dlpi_name, "libapr-1.so", PATH_MAX)) {
#else
    if (strnstr(info->dlpi_name, "libapr-0.so", PATH_MAX)) {
#endif
        void *library = dlopen(info->dlpi_name, RTLD_LAZY);
        if (library) {
            apr_lve_environment_init_fn = dlsym(library, "apr_lve_environment_init");
            apr_lve_environment_init_group_fn = dlsym(library, "apr_lve_environment_init_group");
            apr_lve_environment_init_group_minuid_fn = dlsym(library, "apr_lve_environment_init_group_minuid");
            apr_pool_cleanup_register(p, library, (void *) dlclose, apr_pool_cleanup_null);
            return 1;
        }
    }

    return 0;
}


static void get_lve_func_names (apr_pool_t * p) {
    dl_iterate_phdr (callback, (void *) p);
}

void lve_var_log_config_register (apr_pool_t * p);

static int hostinglimits_pre_config(apr_pool_t * pconf, apr_pool_t * plog, apr_pool_t * ptemp) {
    /* Register us to handle mod_log_config %c/%x variables */
    lve_var_log_config_register(pconf);

    return OK;
}


// function for hook register
static void hostinglimits_module_register_hooks(apr_pool_t * p) {
    int rc = -1;
    errno = 0;
    pw_buff = sysconf(_SC_GETPW_R_SIZE_MAX);
    /* hostinglimits_module_register_hooks called twice */

    if (lve == NULL) {
        if (load_liblve(p) >= 0) {
            lve_pool = p;
            lve = (*init_lve_fn) ((liblve_alloc) lve_alloc, NULL);
            errno_global = errno;
            lve_pool = NULL;

            if (lve != NULL) {
                lve_available = 1;
                // register clean up LVE
                apr_pool_cleanup_register(p, lve, (void *) destroy_lve_fn, apr_pool_cleanup_null);
                get_lve_func_names(p);
                if (apr_lve_environment_init_group_minuid_fn) {
                    found_apr_func_ver = 3;
                } else if (apr_lve_environment_init_group_fn) {
                    found_apr_func_ver = 2;
                } else if (apr_lve_environment_init_fn) {
                    found_apr_func_ver = 1;
                }
            } else {
                lve_available = 0;
            }
        } else {
            lve_available = 0;
        }
    }

    read_cagefs_min_uid();

    static const char *const aszPre[] = {"mod_cache.c", "core.c", "mod_include.c", "mod_php.c", "mod_cgi.c", NULL};

    ap_hook_post_config(hostinglimits_post_config, NULL, NULL, APR_HOOK_REALLY_FIRST);
    ap_hook_pre_config(hostinglimits_pre_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_create_request(hostinglimits_create_request, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(hostinglimits_handler, NULL, aszPre, APR_HOOK_REALLY_FIRST);
}

/*  _________________________________________________________________
**
**  LVE Extension to mod_log_config
**  _________________________________________________________________
*/
#include "mod_log_config.h"

static const char *lve_var_log_handler_y (request_rec * r, char *a);

void lve_var_log_config_register(apr_pool_t * p) {
    static APR_OPTIONAL_FN_TYPE(ap_register_log_handler) * log_pfn_register;

    log_pfn_register = APR_RETRIEVE_OPTIONAL_FN(ap_register_log_handler);

    if (log_pfn_register)
        log_pfn_register(p, "y", lve_var_log_handler_y, 0);

    return;
}

static char *ap_get_lve(request_rec * r) {
    return (char *) apr_table_get(r->notes, X_LVE_ID_HEADER);
}

char *lve_var_lookup(apr_pool_t * p, server_rec * s, conn_rec * c, request_rec * r, char *var) {
    const char *result;
    int resdup;

    result = NULL;

    /*
     * When no pool is given try to find one
     */
    if (p == NULL) {
        if (r != NULL)
            p = r->pool;
        else
            result = NULL;
    }

    /*
     * Request dependent stuff
     */
    if (r != NULL) {
        if (!strcmp(var, "LVE_ID"))
            result = ap_get_lve(r);
    }

    if (result == NULL)
        result = "";
    return (char *) result;
}

static const char *lve_var_log_handler_y(request_rec * r, char *a) {
    char *result = lve_var_lookup(r->pool, r->server, r->connection, r, a);
    if (result != NULL && result[0] == 0)
        result = NULL;

    return result;
}

/**
 * Describing structure of Apache module
 */
#ifdef APLOG_USE_MODULE
APLOG_USE_MODULE(hostinglimits);
#endif

module AP_MODULE_DECLARE_DATA hostinglimits_module = {
    STANDARD20_MODULE_STUFF,
    hostinglimits_module_create_dir_config, /* create directory config */
    hostinglimits_module_merge_config,      /* merging directory config */
    NULL,                                   /* create server config */
    NULL,                                   /* merging server config */
    hostinglimits_module_directives,        /* mapping configuration directives */
    hostinglimits_module_register_hooks     /* registering hooks */
};
